'-------------------------------------------------------------------------------
'
'' Module: keyboard.bas
'
' Copyright 2005-2008 Elba Corp.
' All rights reserved.
'
' Non-exclusive permission to use, copy, modify, and distribute this
' software for any purpose and without fee is hereby granted, provided:
'   1) that the above copyright notice is retained in the source code and,
'   2) that it is understood and agreed that no warranty of any kind,
'      express or implied is provided.
' Use of this code in any manner shall be prima facie evidence of
' acceptance of these terms.
'
' You may contact the author by email at dkinzer@zbasic.net.
'
'-------------------------------------------------------------------------------
'
' Overview:
'
' Matrix keypads are relatively inexpensive and serve well to facilitate
' human input to an embedded system.  A typical embedded system keypad might
' have 12 keys arranged in a 4 row by 3 column matrix.  Each keyswitch connects
' one row line to one column lead.
'
' To interface such a keypad to a microcontroller, it is necessary to have a
' means to activate one row at a time and then examine the column output to
' determine if a key in that row is pressed.  Alternately, the column lines
' can be driven and the row lines observed.
'
' The routines in this module implement the scanning process using a PCF8574
' I/O Expander chip on the I2C bus.  The PCF8574A could be used; the only
' difference between the two chips the the I2C slave address.
'
' This code contains stub routines Beep() and DisplayChar() whose purposes are
' described in their respective headers.  You may want to replace calls to these
' stubs with calls to your equivalent routines.
'
' Lastly, this module contains a simple Main() driver routine intended to be
' used for testing the code.  The Main() is contained within a conditional
' construct so that it can easily be enabled or disabled as necessary.
'
' This code is intended to accompany application note AN205 which contains
' additional information on the hardware connections.
'
'-------------------------------------------------------------------------------

' define the keyboard layout
Private Const kbdRows as Byte = 4    ' number of keypad rows
Private Const kbdCols as Byte = 4    ' number of keypad columns

' delay times, in RTC ticks
Private Const debounceDelay as UnsignedInteger = 15
Private Const autoRepeatDelay As Long = 125
Private Const keyWait As UnsignedInteger = 25

' define some special keys
Private Const KeyEnter as Byte = &H23
Private Const KeyDecimal as Byte = &H2a

' uncomment the line below to use the I2C hardware channel
#define USE_HW_CHAN

#ifdef USE_HW_CHAN
Private Const i2cChan as Byte = 0
#else
Private Const i2cChan as Byte = 1

' These lines define the pin numbers used for the SDA and SCL signals.
' They are needed only if the hardware channel is not used.
#if Option.TargetDevice = "ZX1281" OR Option.TargetDevice = "ZX1280"
Private Const sclPin as Byte = D.0	' to pin 14 of PCF8574A
Private Const sdaPin as Byte = D.1	' to pin 15 of PCF8574A
#else
Private Const sclPin as Byte = C.0	' to pin 14 of PCF8574A
Private Const sdaPin as Byte = C.1	' to pin 15 of PCF8574A
#endif
#endif

' define the address components of the PCF8574
Private Const i2cAddrBase as Byte = &H40	' this part is fixed
Private Const i2cAddrOfst as Byte = &H00	' this part is selectable by A2-A0
Private Const i2cAddr as Byte = i2cAddrBase + i2cAddrOfst

'-------------------------------------------------------------------------------
'
'' KeyboardInit
'
' Initialize the keyboard interface circuitry.
'
Public Sub KeyboardInit()
	Dim data as Byte
	Dim ival as Integer

#ifdef USE_HW_CHAN
	Call OpenI2C(i2cChan, 0, 0)
#else
	Call OpenI2C(i2cChan, sdaPin, sclPin)
#endif

	' set all outputs high
	data = &Hff
	ival = I2CCmd(i2cChan, i2cAddr, 1, data, 0, 0)
End Sub

'-------------------------------------------------------------------------------
'
'' WaitKey
'
' Wait for a key to be pressed and return its code.  If a key is already down
' upon entry, wait for it to be released first.  However, after waiting a
' certain period, auto-repeat is invoked and the key that is down is returned.
'
Public Function WaitKey() as Byte
	Dim key as Byte

	' wait for all keys to be released
	key = KeyIsDown()
	If (key <> 0) Then
		Dim repeatTime as Long

		repeatTime = Register.RTCTick + autoRepeatDelay
		Do
			' wait a bit
	 		Call Sleep(keyWait)

			' see if a key is still down
			If (KeyIsDown() = 0) Then
				' debounce the key release
				Call Sleep(debounceDelay)
				If (KeyIsDown() = 0) Then
					key = 0
					Exit Do
				End If
			ElseIf (Register.RTCTick > repeatTime) Then
				' effect auto-repeat
				Exit Do
			End If
		Loop
	End If

	' wait for a key to be pressed
	Do While (key = 0)
		key = KeyIsDown()
		If (key <> 0) Then
			' debounce the key press
			Call Sleep(debounceDelay)
		Else
	 		Call Sleep(keyWait)
		End If
	Loop

	WaitKey = key
End Function

'-------------------------------------------------------------------------------
'
'' KeyIsDown
'
' Determine if a key is pressed.  If so, return its key code, otherwise, 0.
' Note that this function returns the code of the first key detected.  No
' effort is made to determine if multiple keys are down.
'
Public Function KeyIsDown() as Byte
	Dim row as Byte, col as Byte
	Dim rowBits as Byte, colBits as Byte

	KeyIsDown = 0

	' scan by rows, looking for a key that is down
	rowBits = &Hfe
 	For row = 1 to kbdRows
		Dim ival as Integer

		' set one row line low, read back the column data
		ival = I2CCmd(i2cChan, i2cAddr, 1, rowBits, 1, colBits)

		' scan the column bits looking for a zero
		Const colMask as Byte = Shl(1, kbdRows)
		Dim colBitsSave as Byte
		colBitsSave = colBits
		For col = 1 to kbdCols
			If (Not CBool(colBits And colMask)) Then
 				KeyIsDown = keyCode(row, col)
 				Exit Function
 			End If
			colBits = Shr(colBits, 1)
 		Next
 
 		' prepare to scan the next row
		rowBits = Shl(rowBits, 1) Or &H01
 	Next
End Function

'-------------------------------------------------------------------------------

' This data table maps row/column pairs to ASCII codes.
' Note that it must be indexed as keyTable(col, row).
Private keyTable as ByteTableData({
	"123A"
	"456B"
	"789C"
	"*0#D"
})

'-------------------------------------------------------------------------------
'
'' keyCode
'
' Given row and column values, return the code for the corresponding
' key or zero if the row/column combination is invalid.
'
Private Function keyCode(ByVal row as Byte, ByVal col as Byte) as Byte
	If ((row >= 1) And (row <= CByte(UBound(keyTable, 2))) And _
			(col >= 1) And (col <= CByte(UBound(keyTable, 1)))) Then
		keyCode = keyTable(col, row)
	Else
		keyCode = 0
	End If
End Function

'-------------------------------------------------------------------------------
'
'' GetUnsignedVal
'
' Get an unsigned integer value from the keyboard terminated by an Enter key.  If
' no value is entered before the Enter key is pressed, ignore the Enter key.
'
Public Function GetUnsignedVal() as UnsignedLong
	Dim key as Byte
	Dim val as UnsignedLong
	Dim valChars as Integer

	val = 0
	valChars = 0
	Do
		key = waitKey()
		If (key <> 0) Then
			If (key = KeyEnter) Then
				' "Enter" was pressed
				If (valChars > 0) Then
					Exit Do
				Else
					' no value characters entered yet
					key = 0
				End If
			ElseIf ((key >= &H30) And (key <= &H39)) Then
				' integral digit was pressed
				If ((key <> &H30) Or (val <> 0)) Then
					' add to value
					val = (val * 10) + CULng(key - &H30)
				End If
				valChars = valChars + 1
			Else
				Call Beep()
				key = 0
			End If
			If (key <> 0) then
				DisplayChar(key)
			End If
		End If
	Loop
	GetUnsignedVal = val
End Function

'-------------------------------------------------------------------------------
'
'' GetFloatVal
'
' Get a floating point value from the keyboard terminated by an Enter key.  If
' no value is entered before the Enter key is pressed, ignore the Enter key.
'
Public Function GetFloatVal() as Single
	Dim key as Byte
	Dim val as Single
	Dim fracPart as Long
	Dim fracDigits as Integer
	Dim valChars as Integer

	val = 0.0
	fracDigits = -1
	valChars = 0
	fracPart = 0

	Do
		key = WaitKey()
		If (key <> 0) Then
			If (key = KeyDecimal) Then
				If (fracDigits < 0) Then
					' decimal point was pressed
					fracDigits = 0
					fracPart = 0
					key = &H2e
					valChars = valChars + 1
				Else
					Call Beep()
				End If
			ElseIf (key = KeyEnter) Then
				' "Enter" was pressed
				If (valChars > 0) Then
					' compose value
					Dim fracVal as Single

					If (fracDigits > 0) Then
						fracVal = CSng(fracPart) / (10.0 ^ fracDigits)
					Else
						fracVal = 0.0
					End If
					If ((fracVal <> 0.0) Or (val <> 0.0)) Then
						val = val + fracVal
					End If
					Exit Do
				Else
					' no value characters entered yet
					Key = 0
					Call Beep()
				End If
			ElseIf ((key >= &H30) And (key <= &H39)) Then
				If (fracDigits >= 0) Then
					' fractional digit was pressed
					If (fracDigits < 8) Then
						' add to fraction value
						fracPart = (fracPart * 10) + CLng(key - &H30)
						fracDigits = fracDigits + 1
						valChars = valChars + 1
					Else
						Call Beep()
					End If
				Else
					' integral digit was pressed
					If ((Key <> &H30) Or (Val <> 0.0)) Then
						' add to whole number value
						val = (val * 10.0) + CSng(key - &H30)
					End If
					valChars = valChars + 1
				End If
			Else
				Call Beep()
				key = 0
			End If
			If (key <> 0) then
				DisplayChar(key)
			End If
		End If
	Loop
	GetFloatVal = val
End Function

'-------------------------------------------------------------------------------

'
'' Beep
'
' Make a sound to indicate a keyboard error.
'
Public Sub Beep()
	' add code here to make an appropriate sound
End Sub

'-------------------------------------------------------------------------------

'
'' DisplayChar
'
' Display a character from the keyboard.
'
Public Sub DisplayChar(ByVal ch as Byte)
	' add code here to display a character
	Debug.Print Chr(ch);
End Sub

'-------------------------------------------------------------------------------

' uncomment the next line to include a simple test driver
'#define KEYBOARD_TEST_
#ifdef KEYBOARD_TEST_

Sub Main()
	Call KeyboardInit()

	Dim key as Byte
	Do
		' test code for an individual key
		key = WaitKey()
		Debug.Print CStrHex(key)

		' test code for a Single value
'		Dim f as Single
'		f = GetFloatVal()
'		Debug.Print " --> "; CStr(f)

		' test code for an Integral value
'		Dim v as UnsignedLong
'		v = GetUnsignedVal()
'		Debug.Print " --> "; CStr(v)
	Loop
End Sub
#endif

