'-------------------------------------------------------------------------------
'
'' Module: keyboard.bas
'
' Copyright 2005 Elba Corp.
' All rights reserved.
'
' Non-exclusive permission to use, copy, modify, and distribute this
' software for any purpose and without fee is hereby granted, provided:
'   1) that the above copyright notice is retained in the source code and,
'   2) that it is understood and agreed that no warranty of any kind,
'      express or implied is provided.
' Use of this code in any manner shall be prima facie evidence of
' acceptance of these terms.
'
' You may contact the author by email at dkinzer@zbasic.net.
'
'-------------------------------------------------------------------------------
'
' Overview:
'
' Matrix keypads are relatively inexpensive and serve well to facilitate
' human input to an embedded system.  A typical embedded system keypad might
' have 12 keys arranged in a 4 row by 3 column matrix.  Each keyswitch connects
' one row line to one column lead.
'
' To interface such a keypad to a microcontroller, it is necessary to have a
' means to activate one row at a time and then examine the column output to
' determine if a key in that row is pressed.  Alternately, the column lines
' can be driven and the row lines observed.
'
' The routines in this module implement the scanning process using a shift
' register to drive the row lines and a second shift register to read in
' the column data.  To economize on I/O line usage, both shift registers are
' clocked in common and the data output line to the first shift register is
' used to latch the column data into the second shift register.  This common
' clocking requires that the row driver shift register be reloaded at the
' beginning of each scan loop.  Utilizing separate clock lines would simplify
' the software slightly.  Also, the column outputs could be fed directly into
' the processor, eliminating the shift register.  The modification of the code
' for this arrangement would be relatively straightforward.
'
' This code contains stub routines Beep() and DisplayChar() whose purposes are
' described in their respective headers.  You may want to replace calls to these
' stubs with calls to your equivalent routines.
'
' Lastly, this module contains a simple Main() driver routine intended to be
' used for testing the code.  The Main() is contained within a conditional
' construct so that it can easily be enabled or disabled as necessary.
'
' This code is intended to accompany application note AN205 which contains
' additional information on the hardware connections.
'
'-------------------------------------------------------------------------------

' define the port/pin numbers used
Private Const SR_Clock as Byte = A.1
Private Const SR_DataOut as Byte = A.0

Private Const SR_DataLatch as Byte = A.0
Private Const SR_DataIn as Byte = A.2

' define the keyboard layout
Private Const kbdRows as Byte = 4    ' number of keypad rows
Private Const kbdCols as Byte = 3    ' number of keypad columns

' delay times, in RTC ticks
Private Const debounceDelay as UnsignedInteger = 15
Private Const autoRepeatDelay As Long = 125
Private Const keyWait As UnsignedInteger = 25

' define some special keys
Private Const EnterKey as Byte = &H23
Private Const DecimalKey as Byte = &H2a

'-------------------------------------------------------------------------------
'
'' KeyboardInit
'
' Initialize the keyboard interface circuitry.
'
Public Sub KeyboardInit()
	Call PutPin(SR_Clock, zxOutputLow)
	Call PutPin(SR_DataOut, zxOutputLow)
	Call PutPin(SR_DataIn, zxInputTriState)
End Sub

'-------------------------------------------------------------------------------
'
'' WaitKey
'
' Wait for a key to be pressed and return its code.  If a key is already down
' upon entry, wait for it to be released first.  However, after waiting a
' certain period, auto-repeat is invoked and the key that is down is returned.
'
Public Function WaitKey() as Byte
	Dim key as Byte

	' wait for all keys to be released
	key = KeyIsDown()
	If (key <> 0) Then
		Dim repeatTime as Long

		repeatTime = Register.RTCTick + autoRepeatDelay
		Do
			' wait a bit
	 		Call Sleep(keyWait)

			' see if a key is still down
			If (KeyIsDown() = 0) Then
				' debounce the key release
				Call Sleep(debounceDelay)
				If (KeyIsDown() = 0) Then
					key = 0
					Exit Do
				End If
			ElseIf (Register.RTCTick > repeatTime) Then
				' effect auto-repeat
				Exit Do
			End If
		Loop
	End If

	' wait for a key to be pressed
	Do While (key = 0)
		key = KeyIsDown()
		If (key <> 0) Then
			' debounce the key press
			Call Sleep(debounceDelay)
		Else
	 		Call Sleep(keyWait)
		End If
	Loop

	WaitKey = key
End Function

'-------------------------------------------------------------------------------
'
'' KeyIsDown
'
' Determine if a key is pressed.  If so, return its key code, otherwise, 0.
' Note that this function returns the code of the first key detected.  No
' effort is made to determine if multiple keys are down.
'
Public Function KeyIsDown() as Byte
	Dim row as Byte
	Dim col as Byte
	Dim rowBits as Byte
	Dim colBits as Byte

	KeyIsDown = 0

	' scan by rows, looking for a key that is down
	rowBits = &H10
 	For row = 1 to kbdRows
		' initialize the shift register with one row line high
		Call ShiftOut(SR_DataOut, SR_Clock, 4, rowBits)

		' latch in the column data, shift it in
		Call PutPin(SR_DataLatch, zxOutputLow)
		Call PutPin(SR_DataLatch, zxOutputHigh)
		colBits = ShiftIn(SR_DataIn, SR_Clock, 8)
		
		' scan the column bits looking for a zero
		For col = 1 to kbdCols
			If (Not CBool(colBits And 1)) Then
 				KeyIsDown = keyCode(row, col)
 				Exit Function
 			End If
			colBits = Shr(colBits, 1)
 		Next
 
 		' prepare to scan the next row
		rowBits = Shl(rowBits, 1)
 	Next
End Function

'-------------------------------------------------------------------------------

' This data table maps row/column pairs to ASCII codes.
' Note that it must be indexed as keyTable(col, row).
Private keyTable as ByteTableData({
	"123"
	"456"
	"789"
	"*0#"
})

'-------------------------------------------------------------------------------
'
'' keyCode
'
' Given row and column values, return the code for the corresponding
' key or zero if the row/column combination is invalid.
'
Private Function keyCode(ByVal row as Byte, ByVal col as Byte) as Byte
	If ((row >= 1) And (row <= CByte(UBound(keyTable, 2))) And _
			(col >= 1) And (col <= CByte(UBound(keyTable, 1)))) Then
		keyCode = keyTable(col, row)
	Else
		keyCode = 0
	End if
End Function

'-------------------------------------------------------------------------------
'
'' GetUnsignedVal
'
' Get an unsigned integer value from the keyboard terminated by an Enter key.  If
' no value is entered before the Enter key is pressed, ignore the Enter key.
'
Public Function GetUnsignedVal() as UnsignedLong
	Dim key as Byte
	Dim val as UnsignedLong
	Dim valChars as Integer

	val = 0
	valChars = 0
	Do
		key = waitKey()
		If (key <> 0) Then
			If (key = EnterKey) Then
				' "Enter" was pressed
				If (valChars > 0) Then
					Exit Do
				Else
					' no value characters entered yet
					key = 0
				End If
			ElseIf ((key >= &H30) And (key <= &H39)) Then
				' integral digit was pressed
				If ((key <> &H30) Or (val <> 0)) Then
					' add to value
					val = (val * 10) + CULng(key - &H30)
				End If
				valChars = valChars + 1
			Else
				Call Beep()
				key = 0
			End If
			If (key <> 0) then
				DisplayChar(key)
			End If
		End If
	Loop
	GetUnsignedVal = val
End Function

'-------------------------------------------------------------------------------
'
'' GetFloatVal
'
' Get a floating point value from the keyboard terminated by an Enter key.  If
' no value is entered before the Enter key is pressed, ignore the Enter key.
'
Public Function GetFloatVal() as Single
	Dim key as Byte
	Dim val as Single
	Dim fracPart as Long
	Dim fracDigits as Integer
	Dim valChars as Integer

	val = 0.0
	fracDigits = -1
	valChars = 0
	fracPart = 0

	Do
		key = WaitKey()
		If (key <> 0) Then
			If (key = DecimalKey) Then
				If (fracDigits < 0) Then
					' decimal point was pressed
					fracDigits = 0
					fracPart = 0
					key = &H2e
					valChars = valChars + 1
				Else
					Call Beep()
				End If
			ElseIf (key = EnterKey) Then
				' "Enter" was pressed
				If (valChars > 0) Then
					' compose value
					Dim fracVal as Single

					If (fracDigits > 0) Then
						fracVal = CSng(fracPart) / (10.0 ^ fracDigits)
					Else
						fracVal = 0.0
					End If
					If ((fracVal <> 0.0) Or (val <> 0.0)) Then
						val = val + fracVal
					End If
					Exit Do
				Else
					' no value characters entered yet
					Key = 0
					Call Beep()
				End If
			ElseIf ((key >= &H30) And (key <= &H39)) Then
				If (fracDigits >= 0) Then
					' fractional digit was pressed
					If (fracDigits < 8) Then
						' add to fraction value
						fracPart = (fracPart * 10) + CLng(key - &H30)
						fracDigits = fracDigits + 1
						valChars = valChars + 1
					Else
						Call Beep()
					End If
				Else
					' integral digit was pressed
					If ((Key <> &H30) Or (Val <> 0.0)) Then
						' add to whole number value
						val = (val * 10.0) + CSng(key - &H30)
					End If
					valChars = valChars + 1
				End If
			Else
				Call Beep()
				key = 0
			End If
			If (key <> 0) then
				DisplayChar(key)
			End If
		End If
	Loop
	GetFloatVal = val
End Function

'-------------------------------------------------------------------------------

'
'' Beep
'
' Make a sound to indicate a keyboard error.
'
Public Sub Beep()
	' add code here to make an appropriate sound
End Sub

'-------------------------------------------------------------------------------

'
'' DisplayChar
'
' Display the character whose ASCII value is provided.
'
Public Sub DisplayChar(ByVal ch as Byte)
	' add code here to display a character
	Debug.Print Chr(ch);
End Sub

'-------------------------------------------------------------------------------

' uncomment the next line to include a simple test driver
'#define KEYBOARD_TEST_

#ifdef KEYBOARD_TEST_
Sub Main()
	Call KeyboardInit()

	Dim key as Byte
	Do
		' test code for an individual key
		key = WaitKey()
		Debug.Print CStrHex(key)

		' test code for a Single value
'		Dim f as Single
'		f = GetFloatVal()
'		Debug.Print " --> "; CStr(f)

		' test code for an Integral value
'		Dim v as UnsignedLong
'		v = GetUnsignedVal()
'		Debug.Print " --> "; CStr(v)
	Loop
End Sub
#endif

